import autograd
from autograd import numpy as np
from .sensitivity_lib import _append_jvp
from copy import deepcopy
import warnings


class CrossSensitivity():
    """Calculate a second-order derivative of an optimum with resepct to
    two hyperparameters.

    Given an estimating equation :math:`G(\\theta, \\epsilon_1, \\epsilon_2)`,
    with :math:`G(\\hat\\theta(\\epsilon_1, \\epsilon_2),
    \\epsilon_1, \\epsilon_2) = 0`, this class evaluates a directional
    derivatives

    .. math::
        \\frac{d^2\hat{\\theta}}{d\\epsilon_1 d\\epsilon_2}
        \\Delta \\epsilon_1 \\Delta \\epsilon_2.
    """
    def __init__(self,
                 estimating_equation,
                 solver,
                 input_base,
                 hyper1_base,
                 hyper2_base,
                 term_ii=True,
                 term_i1=True,
                 term_i2=True,
                 term_12=True):

        warnings.warn(
            'The CrossSensitivity class is very experimental and untested.')

        self._g = estimating_equation
        self._solver = solver

        # Copy these because the solver is only valid at these values.
        self._input_base = deepcopy(input_base)
        self._hyper1_base = deepcopy(hyper1_base)
        self._hyper2_base = deepcopy(hyper2_base)

        # TODO: for readability, wrap these so they are always
        # evaluated at the same first three arguments.
        self._g_i = _append_jvp(self._g, num_base_args=3, argnum=0)
        self._g_ii = _append_jvp(self._g_i, num_base_args=3, argnum=0)
        self._g_i1 = _append_jvp(self._g_i, num_base_args=3, argnum=1)
        self._g_i2 = _append_jvp(self._g_i, num_base_args=3, argnum=2)
        self._g_1 = _append_jvp(self._g, num_base_args=3, argnum=1)
        self._g_2 = _append_jvp(self._g, num_base_args=3, argnum=2)
        self._g_12 = _append_jvp(self._g_1, num_base_args=3, argnum=2)

        self._term_ii = term_ii
        self._term_i1 = term_i1
        self._term_i2 = term_i2
        self._term_12 = term_12

    def get_di1(self, dh1):
        g_1 = self._g_1(
            self._input_base, self._hyper1_base, self._hyper2_base,
            dh1)
        di1 = -1 * self._solver(g_1)
        return di1

    def get_di2(self, dh2):
        g_2 = self._g_2(
            self._input_base, self._hyper1_base, self._hyper2_base,
            dh2)
        di2 = -1 * self._solver(g_2)
        return di2

    def evaluate(self, dh1, dh2, di1=None, di2=None, debug=False):
        if self._term_ii or self._term_i2 or self._term_i12:
            if di1 is None:
                di1 = self.get_di1(dh1)

        if self._term_ii or self._term_i1 or self._term_i12:
            if di2 is None:
                di2 = self.get_di2(dh2)

        g_ii = 0
        g_i1 = 0
        g_i2 = 0
        g_12 = 0
        if self._term_ii:
            g_ii = self._g_ii(
                self._input_base, self._hyper1_base, self._hyper2_base,
                di1, di2)

        if self._term_i1:
            g_i1 = self._g_i1(
                self._input_base, self._hyper1_base, self._hyper2_base,
                di2, dh1)

        if self._term_i2:
            g_i2 = self._g_i2(
                self._input_base, self._hyper1_base, self._hyper2_base,
                di1, dh2)

        if self._term_12:
            g_12 = self._g_12(
                self._input_base, self._hyper1_base, self._hyper2_base,
                dh1, dh2)

        if debug:
            print('g_ii: ', g_ii)
            print('g_i1: ', g_i1)
            print('g_i2: ', g_i2)
            print('g_12: ', g_12)
            print('di1: ', di1)
            print('di2: ', di2)

        return -1 * self._solver(g_ii + g_i1 + g_i2 + g_12)


class OptimumChecker():
    def __init__(self,
                 estimating_equation,
                 solver,
                 input_base,
                 hyper_base):
        """Estimate the error in sensitivity due to incomplete optimization.

        Parameters
        ---------------
        estimating_equation : `callable`
            A function taking arguments `(input, hyper)` and returning
            a vector, typically the same length as the input.  The idea is
            that `estimating_equation(input_base, hyper_base) = [0, ..., 0]`.
        solver : `callable`
            A function of a single vector variable `v` solving :math:`H^{-1} v`,
            where `H` is the Hessian of the estimating equation with respect
            to the input variable at `input_base`, `hyper_base`.
        input_base : `numpy.ndarray`
            The base value of the parameter to be optimized
        hyper_base : `numpy.ndarray`
            The base value of the hyperparameter.
        """

        self._input_base = deepcopy(input_base)
        self._hyper_base = deepcopy(hyper_base)
        self._solver = solver

        def estimating_equation_lagrange(ipar, hpar, lam):
            return estimating_equation(ipar, hpar) + lam
        self.estimating_equation_lagrange = estimating_equation_lagrange

        # self._obj = lambda ipar: estimating_equation(ipar, self._hyper_base)
        # self._obj_grad = autograd.jacobian(self._obj)

        #self._lam_base = -1 * self._obj_grad(self._input_base)
        self._lam_base = \
            -1 * estimating_equation(self._input_base, self._hyper_base)
        self._dlam = -1 * self._lam_base

        self._cross_sens = CrossSensitivity(
            estimating_equation = self.estimating_equation_lagrange,
            solver =        self._solver,
            input_base =    self._input_base,
            hyper1_base =   self._hyper_base,
            hyper2_base =   self._lam_base,
            term_i2=False,
            term_12=False)

    def get_newton_step(self):
        """Return a Netwon step towards the optimum.
        """
        return self._cross_sens.get_di2(self._dlam)

    def get_dinput_dhyper(self, dhyper):
        """Return the first directional derivative of the optimum with respect
        to the hyperparameter in the direction `dhyper`.
        """
        return self._cross_sens.get_di1(dhyper)

    def correction(self, hyper_new, dinput_dhyper=None, newton_step=None):
        """Return the first-order correction to the change in
        dinput_dhyper as you take a Newton step.
        """
        dhyper = hyper_new - self._hyper_base
        if dinput_dhyper is None:
            dinput_dhyper = self.get_dinput_dhyper(dhyper)
        if newton_step is None:
            newton_step = self.get_newton_step()
        dinput_dhyper_correction = \
            self._cross_sens.evaluate(
                dhyper, self._dlam,
                di1=dinput_dhyper,
                di2=newton_step)
        return dinput_dhyper_correction
        return self._input_base + dinput_dhyper + dinput_dhyper_correction

    def evaluate(self, hyper_new, dinput_dhyper=None, newton_step=None):
        """Return the first-order approximation to the change in
        dinput_dhyper as you take a Newton step.
        """
        dhyper = hyper_new - self._hyper_base
        if dinput_dhyper is None:
            dinput_dhyper = self.get_dinput_dhyper(dhyper)
        dinput_dhyper_correction = self.correction(
            hyper_new, dinput_dhyper=dinput_dhyper, newton_step=newton_step)
        return self._input_base + dinput_dhyper + dinput_dhyper_correction



from vittles.sensitivity_lib import \
    HyperparameterSensitivityLinearApproximation, \
    ParametricSensitivityTaylorExpansion

from vittles.sparse_hessian_lib import SparseBlockHessian
from vittles.lr_cov_lib import LinearResponseCovariances
import vittles.solver_lib
import vittles.bivariate_sensitivity_lib

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import scipy as sp
import scipy.sparse
from scipy.linalg import cho_factor, cho_solve
import warnings


def get_dense_cholesky_solver(h, h_chol=None):
    """Return a function solving :math:`h^{-1} v` with a dense
    Cholesky factorization.

    Parameters
    -------------
    h : `numpy.ndarray`
        A dense symmetric positive definite matrix.
    h_chol : A Cholesky factorization
        Optional, a Cholesky factorization created with `sp.linalg.cho_factor`.
        If this is specified, the argument `h` is ignored.  If `None` (the
        default), the factoriztion of `h` is calculated.

    Returns
    --------------
    solve : `callable`
        A function of a single vector argument, `v` that returns
        :math:`h^{-1} v`.
    """
    if h_chol is None:
        h_chol = sp.linalg.cho_factor(h)
    def solve(v):
        return sp.linalg.cho_solve(h_chol, v)
    return solve


def get_sparse_cholesky_solver(h):
    """Return a function solving :math:`h^{-1} v` for a sparse `h`.

    Parameters
    -------------
    h : A sparse invertible matrix.

    Returns
    --------------
    solve : `callable`
        A function of a single vector argument, `v` that returns
        :math:`h^{-1} v`.
    """
    if not sp.sparse.issparse(h):
        raise ValueError('`h` must be sparse.')
    return sp.sparse.linalg.factorized(h)


def get_cholesky_solver(h):
    """Return a function solving :math:`h^{-1} v` for a matrix `h`.

    Parameters
    -------------
    h : A dense or sparse matrix.

    Returns
    --------------
    solve : `callable`
        A function of a single vector argument, `v` that returns
        :math:`h^{-1} v`.
    """
    if sp.sparse.issparse(h):
        return get_sparse_cholesky_solver(h)
    else:
        return get_dense_cholesky_solver(h)


def get_cg_solver(mat_times_vec, dim, cg_opts={}):
    """Return a function solving :math:`h^{-1} v` for a matrix `h` using
    conjugate gradient.

    Parameters
    -------------
    mat_times_vec : `callable`
        A function of a single vector argument, `v` that returns the product
        `h v` for some invertible matrix `h`.
    dim : `int`
        The dimension of the vector `v`.
    cg_opts : `dict`
        Optional, a dictionary of named arguments for the solver
        `sp.sparse.linalg.cg`.

    Returns
    ---------
    solve : `callable`
        A function of a single vector argument, `v` that returns the
        conjugate gradient approximation to :math:`h^{-1} v`.
    """
    linop = sp.sparse.linalg.LinearOperator((dim, dim), mat_times_vec)
    def solve(v):
        cg_result = sp.sparse.linalg.cg(linop, v, **cg_opts, atol='legacy')
        if cg_result[1] != 0:
            warnings.warn(
                'CG exited with error code {}'.format(cg_result[1]))
        return cg_result[0]
    return solve



##############
# LRVB class #
##############

import autograd
from copy import deepcopy
import numpy as np
from . import solver_lib

class LinearResponseCovariances:
    """
    Calculate linear response covariances of a variational distribution.

    Let :math:`q(\\theta | \\eta)` be a class of probability distribtions on
    :math:`\\theta` where the class is parameterized by the real-valued vector
    :math:`\\eta`.  Suppose that we wish to approximate a distribution
    :math:`q(\\theta | \\eta^*) \\approx p(\\theta)` by solving an optimization
    problem :math:`\\eta^* = \\mathrm{argmin} f(\\eta)`.  For example, :math:`f`
    might be a measure of distance between :math:`q(\\theta | \\eta)` and
    :math:`p(\\theta)`.  This class uses the sensitivity of the optimal
    :math:`\\eta^*` to estimate the covariance
    :math:`\\mathrm{Cov}_p(g(\\theta))`. This covariance estimate is called the
    "linear response covariance".

    In this notation, the arguments to the class mathods are as follows.
    :math:`f` is ``objective_fun``, :math:`\\eta^*` is ``opt_par_value``, and
    the function ``calculate_moments`` evaluates :math:`\\mathbb{E}_{q(\\theta |
    \\eta)}[g(\\theta)]` as a function of :math:`\\eta`.

    Methods
    ------------
    set_base_values:
        Set the base values, :math:`\\eta^*` that optimizes the
        objective function.
    get_hessian_at_opt:
        Return the Hessian of the objective function evaluated at the optimum.
    get_hessian_cholesky_at_opt:
        Return the Cholesky decomposition of the Hessian of the objective
        function evaluated at the optimum.
    get_lr_covariance:
        Return the linear response covariance of a given moment.
    """
    def __init__(
        self,
        objective_fun,
        opt_par_value,
        validate_optimum=False,
        hessian_at_opt=None,
        factorize_hessian=True,
        grad_tol=1e-8):
        """
        Parameters
        --------------
        objective_fun: Callable function
            A callable function whose optimum parameterizes an approximate
            Bayesian posterior.  The function must take as a single
            argument a numeric vector, ``opt_par``.
        opt_par_value:
            The value of ``opt_par`` at which ``objective_fun`` is optimized.
        validate_optimum: Boolean
            When setting the values of ``opt_par``, check
            that ``opt_par`` is, in fact, a critical point of
            ``objective_fun``.
        hessian_at_opt: Numeric matrix (optional)
            The Hessian of ``objective_fun`` at the optimum.  If not specified,
            it is calculated using automatic differentiation.
        factorize_hessian: Boolean
            If ``True``, solve the required linear system using a Cholesky
            factorization.  If ``False``, use the conjugate gradient algorithm
            to avoid forming or inverting the Hessian.
        grad_tol: Float
            The tolerance used to check that the gradient is approximately
            zero at the optimum.
        """

        self._obj_fun = objective_fun
        self._obj_fun_grad = autograd.grad(self._obj_fun, argnum=0)
        self._obj_fun_hessian = autograd.hessian(self._obj_fun, argnum=0)
        self._obj_fun_hvp = autograd.hessian_vector_product(
            self._obj_fun, argnum=0)

        self._grad_tol = grad_tol

        self.set_base_values(
            opt_par_value, hessian_at_opt,
            factorize_hessian, validate=validate_optimum)

    def set_base_values(self,
                        opt_par_value,
                        hessian_at_opt,
                        factorize_hessian=True,
                        validate=True,
                        grad_tol=None):
        if grad_tol is None:
            grad_tol = self._grad_tol

        # Set the values of the optimal parameters.
        self._opt0 = deepcopy(opt_par_value)

        # Set the values of the Hessian at the optimum.
        if hessian_at_opt is None:
            self._hess0 = self._obj_fun_hessian(self._opt0)
        else:
            self._hess0 = hessian_at_opt

        self.hess_solver = solver_lib.get_cholesky_solver(self._hess0)

        if validate:
            # Check that the gradient of the objective is zero at the optimum.
            grad0 = self._obj_fun_grad(self._opt0)
            newton_step = -1 * self.hess_solver(grad0)

            newton_step_norm = np.linalg.norm(newton_step)
            if newton_step_norm > grad_tol:
                err_msg = \
                    'The gradient is not zero at the proposed optimal ' + \
                    'values.  ||newton_step|| = {} > {} = grad_tol'.format(
                        newton_step_norm, grad_tol)
                raise ValueError(err_msg)

    def get_hessian_at_opt(self):
        return self._hess0

    def get_lr_covariance_from_jacobians(self,
                                         moment_jacobian1,
                                         moment_jacobian2):
        """
        Get the linear response covariance between two vectors of moments.

        Parameters
        ------------
        moment_jacobian1: 2d numeric array.
            The Jacobian matrix of a map from a value of
            ``opt_par`` to a vector of moments of interest.  Following
            standard notation for Jacobian matrices, the rows should
            correspond to moments and the columns to elements of
            a flattened ``opt_par``.
        moment_jacobian2: 2d numeric array.
            Like ``moment_jacobian1`` but for the second vector of moments.

        Returns
        ------------
        Numeric matrix
            If ``moment_jacobian1(opt_par)`` is the Jacobian
            of :math:`\mathbb{E}_q[g_1(\\theta)]` and
            ``moment_jacobian2(opt_par)``
            is the Jacobian of  :math:`\mathbb{E}_q[g_2(\\theta)]` then this
            returns the linear response estimate of
            :math:`\\mathrm{Cov}_p(g_1(\\theta), g_2(\\theta))`.
        """

        if moment_jacobian1.ndim != 2:
            raise ValueError('moment_jacobian1 must be a 2d array.')

        if moment_jacobian2.ndim != 2:
            raise ValueError('moment_jacobian2 must be a 2d array.')

        if moment_jacobian1.shape[1] != len(self._opt0):
            err_msg = ('The number of rows of moment_jacobian1 must match' +
                       'the dimension of the optimization parameter. ' +
                       'Expected {} rows, but got shape = {}').format(
                         len(self._opt0), moment_jacobian1.shape)
            raise ValueError(err_msg)

        if moment_jacobian2.shape[1] != len(self._opt0):
            err_msg = ('The number of rows of moment_jacobian2 must match' +
                       'the dimension of the optimization parameter. ' +
                       'Expected {} rows, but got shape = {}').format(
                         len(self._opt0), moment_jacobian2.shape)
            raise ValueError(err_msg)

        return moment_jacobian1 @ self.hess_solver(moment_jacobian2.T)

    def get_moment_jacobian(self, calculate_moments):
        """
        The Jacobian matrix of a map from ``opt_par`` to a vector of
        moments of interest.

        Parameters
        ------------
        calculate_moments: Callable function
            A function that takes the folded ``opt_par`` as a single argument
            and returns a numeric vector containing posterior moments of
            interest.

        Returns
        ----------
        Numeric matrix
            The Jacobian of the moments.
        """
        calculate_moments_jacobian = autograd.jacobian(calculate_moments)
        return calculate_moments_jacobian(self._opt0)

    def get_lr_covariance(self, calculate_moments):
        """
        Get the linear response covariance of a vector of moments.

        Parameters
        ------------
        calculate_moments: Callable function
            A function that takes the folded ``opt_par`` as a single argument
            and returns a numeric vector containing posterior moments of
            interest.

        Returns
        ------------
        Numeric matrix
            If ``calculate_moments(opt_par)`` returns
            :math:`\\mathbb{E}_q[g(\\theta)]`
            then this returns the linear response estimate of
            :math:`\\mathrm{Cov}_p(g(\\theta))`.
        """

        moment_jacobian = self.get_moment_jacobian(calculate_moments)
        return self.get_lr_covariance_from_jacobians(
            moment_jacobian, moment_jacobian)



##########################################################################
# Functions for evaluating the sensitivity of optima to hyperparameters. #
##########################################################################

import autograd
import autograd.numpy as np
from autograd.core import primitive, defvjp, defjvp

from copy import deepcopy
from math import factorial
from string import ascii_lowercase
import warnings

from paragami import FlattenFunctionInput
from . import solver_lib


class HyperparameterSensitivityLinearApproximation:
    """
    Linearly approximate dependence of an optimum on a hyperparameter.

    Suppose we have an optimization problem in which the objective
    depends on a hyperparameter:

    .. math::

        \hat{\\theta} = \mathrm{argmin}_{\\theta} f(\\theta, \\lambda).

    The optimal parameter, :math:`\hat{\\theta}`, is a function of
    :math:`\\lambda` through the optimization problem.  In general, this
    dependence is complex and nonlinear.  To approximate this dependence,
    this class uses the linear approximation:

    .. math::

        \hat{\\theta}(\\lambda) \\approx \hat{\\theta}(\\lambda_0) +
            \\frac{d\hat{\\theta}}{d\\lambda}|_{\\lambda_0}
                (\\lambda - \\lambda_0).

    In terms of the arguments to this function,
    :math:`\\theta` corresponds to ``opt_par``,
    :math:`\\lambda` corresponds to ``hyper_par``,
    and :math:`f` corresponds to ``objective_fun``.

    Methods
    ------------
    set_base_values:
        Set the base values, :math:`\\lambda_0` and
        :math:`\\theta_0 := \hat\\theta(\\lambda_0)`, at which the linear
        approximation is evaluated.
    get_dopt_dhyper:
        Return the Jacobian matrix
        :math:`\\frac{d\hat{\\theta}}{d\\lambda}|_{\\lambda_0}` in flattened
        space.
    get_hessian_at_opt:
        Return the Hessian of the objective function in the
        flattened space.
    predict_opt_par_from_hyper_par:
        Use the linear approximation to predict
        the value of ``opt_par`` from a value of ``hyper_par``.
    """
    def __init__(
        self,
        objective_fun,
        opt_par_value,
        hyper_par_value,
        validate_optimum=False,
        hessian_at_opt=None,
        cross_hess_at_opt=None,
        hyper_par_objective_fun=None,
        grad_tol=1e-8):
        """
        Parameters
        --------------
        objective_fun : `callable`
            The objective function taking two positional arguments,
            - ``opt_par``: The parameter to be optimized (`numpy.ndarray` (N,))
            - ``hyper_par``: A hyperparameter (`numpy.ndarray` (N,))
            and returning a real value to be minimized.
        opt_par_value :  `numpy.ndarray` (N,)
            The value of ``opt_par`` at which ``objective_fun`` is
            optimized for the given value of ``hyper_par_value``.
        hyper_par_value : `numpy.ndarray` (M,)
            The value of ``hyper_par`` at which ``opt_par`` optimizes
            ``objective_fun``.
        validate_optimum : `bool`, optional
            When setting the values of ``opt_par`` and ``hyper_par``, check
            that ``opt_par`` is, in fact, a critical point of
            ``objective_fun``.
        hessian_at_opt : `numpy.ndarray` (N,N), optional
            The Hessian of ``objective_fun`` at the optimum.  If not specified,
            it is calculated using automatic differentiation.
        cross_hess_at_opt : `numpy.ndarray`  (N, M)
            Optional.  The second derivative of the objective with respect to
            ``input_val`` then ``hyper_val``.
            If not specified it is calculated at initialization.
        hyper_par_objective_fun : `callable`, optional
            The part of ``objective_fun`` depending on both ``opt_par`` and
            ``hyper_par``.  The arguments must be the same as
            ``objective_fun``:
            - ``opt_par``: The parameter to be optimized (`numpy.ndarray` (N,))
            - ``hyper_par``: A hyperparameter (`numpy.ndarray` (N,))
            This can be useful if only a small part of the objective function
            depends on both ``opt_par`` and ``hyper_par``.  If not specified,
            ``objective_fun`` is used.
        grad_tol : `float`, optional
            The tolerance used to check that the gradient is approximately
            zero at the optimum.
        """

        self._objective_fun = objective_fun
        self._obj_fun_grad = autograd.grad(self._objective_fun, argnum=0)
        self._obj_fun_hessian = autograd.hessian(self._objective_fun, argnum=0)
        self._obj_fun_hvp = autograd.hessian_vector_product(
            self._objective_fun, argnum=0)

        if hyper_par_objective_fun is None:
            self._hyper_par_objective_fun = self._objective_fun
        else:
            self._hyper_par_objective_fun = hyper_par_objective_fun

        # TODO: is this the right default order?  Make this flexible.
        self._hyper_obj_fun_grad = \
            autograd.grad(self._hyper_par_objective_fun, argnum=0)
        self._hyper_obj_cross_hess = autograd.jacobian(
            self._hyper_obj_fun_grad, argnum=1)

        self._grad_tol = grad_tol

        self.set_base_values(
            opt_par_value, hyper_par_value,
            hessian_at_opt, cross_hess_at_opt,
            validate_optimum=validate_optimum,
            grad_tol=self._grad_tol)

    def set_base_values(self,
                        opt_par_value, hyper_par_value,
                        hessian_at_opt, cross_hess_at_opt,
                        validate_optimum=True, grad_tol=None):

        # Set the values of the optimal parameters.
        self._opt0 = deepcopy(opt_par_value)
        self._hyper0 = deepcopy(hyper_par_value)

        # Set the values of the Hessian at the optimum.
        if hessian_at_opt is None:
            self._hess0 = self._obj_fun_hessian(self._opt0, self._hyper0)
        else:
            self._hess0 = hessian_at_opt
        if self._hess0.shape != (len(self._opt0), len(self._opt0)):
            raise ValueError('``hessian_at_opt`` is the wrong shape.')

        self.hess_solver = solver_lib.get_cholesky_solver(self._hess0)

        if validate_optimum:
            if grad_tol is None:
                grad_tol = self._grad_tol

            # Check that the gradient of the objective is zero at the optimum.
            grad0 = self._obj_fun_grad(self._opt0, self._hyper0)
            newton_step = -1 * self.hess_solver(grad0)

            newton_step_norm = np.linalg.norm(newton_step)
            if newton_step_norm > grad_tol:
                err_msg = \
                    'The gradient is not zero at the proposed optimal ' + \
                    'values.  ||newton_step|| = {} > {} = grad_tol'.format(
                        newton_step_norm, grad_tol)
                raise ValueError(err_msg)

        if cross_hess_at_opt is None:
            self._cross_hess = self._hyper_obj_cross_hess(self._opt0, self._hyper0)
        else:
            self._cross_hess = cross_hess_at_opt
        if self._cross_hess.shape != (len(self._opt0), len(self._hyper0)):
            raise ValueError('``cross_hess_at_opt`` is the wrong shape.')

        self._sens_mat = -1 * self.hess_solver(self._cross_hess)


    # Methods:
    def get_dopt_dhyper(self):
        return self._sens_mat

    def get_hessian_at_opt(self):
        return self._hess0

    def predict_opt_par_from_hyper_par(self, new_hyper_par_value):
        """
        Predict ``opt_par`` using the linear approximation.

        Parameters
        ------------
        new_hyper_par_value: `numpy.ndarray` (M,)
            The value of ``hyper_par`` at which to approximate ``opt_par``.
        """
        return \
            self._opt0 + \
            self._sens_mat @ (new_hyper_par_value - self._hyper0)

    def _check_hyper_par_value(self, hyper_par, tolerance=1e-8):
        if np.max(np.abs(hyper_par - self._hyper0)) > tolerance:
            raise ValueError(
                'get_opt_par must be evaluated at ',
                self._hyper0, ' != ', hyper_par)

    def get_opt_par_function(self):
        """Return a differentiable function returning the optimal value.
        """

        @primitive
        def get_opt_par(hyper_par):
            """Evaluate the optimum as a differentiable function.

            Parameters
            -------------
            hyper_par: numeric
                The value must match `hyper_par_value`, but it can be an
                array box, allowing functions of the optimal parameter to be
                differentiated with `autograd`.
            """

            self._check_hyper_par_value(hyper_par)
            return self._opt0

        # Reverse mode.
        def _get_parhat_vjp(ans, hyperpar):
            self._check_hyper_par_value(hyperpar)

            # Make this primitive to prevent attempting higher-order derivatives.
            # To do so efficiently will require a higher order approximation.
            @primitive
            def vjp(g):
                return g.T @ self._sens_mat
            return vjp

        # Forward mode.
        # Make this primitive to prevent attempting higher-order derivatives.
        # To do so efficiently will require a higher order approximation.
        @primitive
        def _get_parhat_jvp(g, ans, hyperpar):
            self._check_hyper_par_value(hyperpar)
            return self._sens_mat @ g

        # Define the derivatives of `get_opt_par`
        defvjp(get_opt_par, _get_parhat_vjp)
        defjvp(get_opt_par, _get_parhat_jvp)

        return get_opt_par

################################
# Higher-order approximations. #
################################

def _append_jvp(fun, num_base_args=1, argnum=0):
    """
    Append a jacobian vector product to a function.

    This function is designed to be used recursively to calculate
    higher-order Jacobian-vector products.

    Parameters
    --------------
    fun: Callable function
        The function to be differentiated.
    num_base_args: integer
        The number of inputs to the base function, i.e.,
        to the function before any differentiation.
     argnum: inteeger
        Which argument should be differentiated with respect to.
        Must be between 0 and num_base_args - 1.

    Returns
    ------------
    Denote the base args x1, ..., xB, where B == num_base_args.
    Let argnum = k.  Then _append_jvp returns a function,
    fun_jvp(x1, ..., xB, ..., v) =
    \sum_i (dfun_dx_{ki}) v_i | (x1, ..., xB).
    That is, it returns the Jacobian vector product where the Jacobian
    is taken with respect to xk, and the vector product is with the
    final argument.
    """
    assert argnum < num_base_args

    fun_jvp = autograd.make_jvp(fun, argnum=argnum)
    def obj_jvp_wrapper(*argv):
        # These are the base arguments -- the points at which the
        # Jacobians are evaluated.
        base_args = argv[0:num_base_args]

        # The rest of the arguments are the vectors, with which inner
        # products are taken in the order they were passed to
        # _append_jvp.
        vec_args = argv[num_base_args:]

        if (len(vec_args) > 1):
            # Then this is being applied to an existing Jacobian
            # vector product.  The last will be the new vector, and
            # we need to evaluate the function at the previous vectors.
            # The new jvp will be appended to the end of the existing
            # list.
            old_vec_args = vec_args[:-1]
            return fun_jvp(*base_args, *old_vec_args)(vec_args[-1])[1]
        else:
            return fun_jvp(*base_args)(*vec_args)[1]

    return obj_jvp_wrapper


class DerivativeTerm:
    """
    A single term in a Taylor expansion of a two-parameter objective with
    methods for computing its derivatives.

    .. note::
        This class is intended for internal use.  Most users should not
        use ``DerivativeTerm`` directly, and should rather use
        ``ParametricSensitivityTaylorExpansion``.

    Let :math:`\hat{\\eta}(\\epsilon)` be such that
    :math:`g(\hat{\\eta}(\\epsilon), \\epsilon) = 0`.
    The nomenclature assumes that
    the term arises from calculating total derivatives of
    :math:`g(\hat{\\eta}(\\epsilon), \\epsilon)`,
    with respect to :math:`\\epsilon`, so such a term arose from repeated
    applications of the chain and product rule of differentiation with respect
    to :math:`\\epsilon`.

    In the ``ParametricSensitivityTaylorExpansion`` class, such terms are
    then used to calculate

    .. math::
        \\frac{d^k\hat{\\eta}}{d\\epsilon^k} |_{\\eta_0, \\epsilon_0}.

    We assume the term will only be calculated summed against a single value
    of :math:`\\Delta\\epsilon`, so we do not need to keep track of the
    order in which the derivatives are evaluated.

    Every term arising from differentiation of :math:`g(\hat{\\eta}(\\epsilon),
    \\epsilon)` with respect to :math:`\\epsilon` is a product the following
    types of terms.

    First, there are the partial derivatives of :math:`g` itself.

    .. math::
        \\frac{\\partial^{m+n} g(\\eta, \\epsilon)}
              {\\partial \\eta^m \\epsilon^n}

    In the preceding display, ``m``
    is the total number of :math:`\\eta` derivatives, i.e.
    ``m = np.sum(eta_orders)``, and ``n = eps_order``.

    Each partial derivative of :math:`g` with respect to :math:`\\epsilon`
    will multiply one :math:`\\Delta \\epsilon` term directly.  Each
    partial derivative with respect to :math:`\\eta` will multiply a term
    of the form

    .. math::
        \\frac{d^p \hat{\\eta}}{d \\epsilon^p}

    which will in turn multiply :math:`p` different :math:`\\Delta \\epsilon`
    terms. The number of such terms of order :math:`p` are given by the entry
    ``eta_orders[p - 1]``.  Each such terms arises from a single partial
    derivative of :math:`g` with respect to :math:`\\eta`, which is why
    the above ``m = np.sum(eta_orders)``.

    Finally, the term is multiplied by the constant ``prefactor``.

    For example, suppose that ``eta_orders = [1, 0, 2]``, ``prefactor = 1.5``,
    and ``epsilon_order = 2``.  Then the derivative term is

    .. math::
        1.5 \\cdot
        \\frac{\\partial^{5} g(\hat{\\eta}, \\epsilon)}
              {\\partial \\eta^3 \\epsilon^2} \\cdot
        \\frac{d \hat{\\eta}}{d \\epsilon} \\cdot
        \\frac{d^3 \hat{\\eta}}{d \\epsilon^3} \\cdot
        \\frac{d^3 \hat{\\eta}}{d \\epsilon^3} \\cdot

    ...which will multiply a total of
    ``9 = epsilon_order + np.sum(eta_orders * [1, 2, 3])``
    :math:`\\Delta \\epsilon` terms.  Such a term would arise in
    the 9-th order Taylor expansion of :math:`g(\hat{\\eta}(\\epsilon),
    \\epsilon)` in :math:`\\epsilon`.

    Attributes
    -----------------
    eps_order:
        The total number of epsilon derivatives of g.
    eta_orders:
        A vector of length order - 1.  Entry i contains the number
        of terms of the form d\eta^{i + 1} / d\epsilon^{i + 1}.
    prefactor:
        The constant multiple in front of this term.

    Methods
    ------------
    differentiate:
        Get a list of derivatives terms resulting from differentiating this
        term.
    check_similarity:
        Return a boolean indicating whether this term is equivalent to another
        term in the order of its derivative.
    combine_with:
        Return the sum of this term and another term.
    """
    def __init__(self, eps_order, eta_orders, prefactor):
        """
        Parameters
        -------------
        eps_order:
            The total number of epsilon partial derivatives of g.
        eta_orders:
            A vector of length order - 1.  Entry i contains the number
            of terms :math:`d\\eta^{i + 1} / d\\epsilon^{i + 1}`.
        prefactor:
            The constant multiple in front of this term.
        """
        # Base properties.
        self.eps_order = eps_order
        self.eta_orders = eta_orders
        self.prefactor = prefactor

        # Derived quantities.

        # The total number of eta partials.
        self.total_eta_order = np.sum(self.eta_orders)

        # The order is the total number of epsilon derivatives.
        self._order = int(
            self.eps_order + \
            np.sum(self.eta_orders * np.arange(1, len(self.eta_orders) + 1)))

        # Sanity checks.
        # The rules of differentiation require that these assertions be true
        # -- that is, if terms are generated using the differentiate()
        # method from other well-defined terms, these assertions should always
        # be sastisfied.
        assert isinstance(self.eps_order, int)
        assert len(self.eta_orders) == self._order
        assert self.eps_order >= 0 # Redundant
        for eta_order in self.eta_orders:
            assert eta_order >= 0
            assert isinstance(eta_order, int)

    def __str__(self):
        return 'Order: {}\t{} * eta{} * eps[{}]'.format(
            self._order, self.prefactor, self.eta_orders, self.eps_order)

    def order(self):
        return self._order

    def differentiate(self):
        derivative_terms = []

        old_eta_orders = deepcopy(self.eta_orders)
        old_eta_orders.append(0)

        # dG / deps.
        derivative_terms.append(
            DerivativeTerm(
                eps_order=self.eps_order + 1,
                eta_orders=deepcopy(old_eta_orders),
                prefactor=self.prefactor))

        # dG / deta.
        new_eta_orders = deepcopy(old_eta_orders)
        new_eta_orders[0] = new_eta_orders[0] + 1
        derivative_terms.append(
            DerivativeTerm(
                eps_order=self.eps_order,
                eta_orders=new_eta_orders,
                prefactor=self.prefactor))

        # Derivatives of each d^{i}eta / deps^i term.
        for i in range(len(self.eta_orders)):
            eta_order = self.eta_orders[i]
            if eta_order > 0:
                new_eta_orders = deepcopy(old_eta_orders)
                new_eta_orders[i] = new_eta_orders[i] - 1
                new_eta_orders[i + 1] = new_eta_orders[i + 1] + 1
                derivative_terms.append(
                    DerivativeTerm(
                        eps_order=self.eps_order,
                        eta_orders=new_eta_orders,
                        prefactor=self.prefactor * eta_order))

        return derivative_terms

    # Return whether another term matches this one in the pattern of derivatives.
    def check_similarity(self, term):
        return \
            (self.eps_order == term.eps_order) & \
            (self.eta_orders == term.eta_orders)

    # Assert that another term has the same pattern of derivatives and
    # return a new term that combines the two.
    def combine_with(self, term):
        assert self.check_similarity(term)
        return DerivativeTerm(
            eps_order=self.eps_order,
            eta_orders=self.eta_orders,
            prefactor=self.prefactor + term.prefactor)


def _evaluate_term_fwd(term, eta0, eps0, deps, eta_derivs,
                       eval_directional_derivative, validate=False):
    """Evaluate the DerivativeTerm in forward mode.

    Parameters
    ----------------------
    term: `DerivativeTerm`
        A `DerivativeTerm` object to evaluate.
    eta0, eps0 : `numpy.ndarray`
        Where to evaluate the derivative.
    deps : `numpy.ndarray`
        The direction in which to evaluate the derivative.
    eta_derivs : `list` of `numpy.ndarray`
        A list where ``eta_derivs[i]`` contains
        :math:`d\\eta^i / d\\epsilon^i \\Delta \\epsilon^i`.
    eval_directional_derivative:
        A function matching `eval_directional_derivative` in
        ForwardModeDerivativeArray.
    validate: optional `bool`
        If `True`, run checks for the appropriate sizes of arguments and
        produce helpful error messages if necessary.
        Default is `False`.
    """

    if validate:
        if len(eta_derivs) < term.order() - 1:
            raise ValueError('Not enough derivatives in ``eta_derivs``.')

    # First eta arguments, then epsilons.
    vec_args = []

    eta_directions = []
    for i in range(len(term.eta_orders)):
        eta_order = term.eta_orders[i]
        if eta_order > 0:
            for j in range(eta_order):
                eta_directions.append(eta_derivs[i])

    eps_directions = []
    for i in range(term.eps_order):
        eps_directions.append(deps)

    return term.prefactor * eval_directional_derivative(
        eta0, eps0, eta_directions, eps_directions, validate=validate)


def _contract_tensor(deriv_array, dx1s, dx2s):
    """Apply deriv_array to the list of vectors in dx1s and dx2s, keeping the
    first dimension of deriv_array.

    This is best understood with an example.  Consider:

    >>> deriv_array = np.random.random((3, 4, 4, 2))
    >>> dx1s = [ np.random.random(4) for _ in range(2) ]
    >>> dx2s = [ np.random.random(2) for _ in range(1) ]

    We want
    >>> einsum('zabc,a,b,c->z', deriv_array, dx1s[0], dx1s[1], dx2s[0])

    We construct this automatically for any length vectors.
    """
    if len(dx1s) + len(dx2s) >= len(ascii_lowercase):
        raise ValueError(
            'You cannot use _contract_tensor with so many vectors ' +
            'because of the crazy way rgiordan decided to do this with einsum.')

    if (len(dx1s) + len(dx2s)) == 0:
        return deriv_array

    ind_letters =  ascii_lowercase[0:(len(dx1s) + len(dx2s))]
    einsum_str = \
        'z' + ''.join(ind_letters) + ',' + ','.join(ind_letters) + '->z'
    return np.einsum(einsum_str, deriv_array, *(dx1s + dx2s))


class ForwardModeDerivativeArray():
    def __init__(self, fun, order1, order2):
        self._order1 = order1
        self._order2 = order2

        self._eval_fun_derivs = [[ fun ]]
        for x1_ind in range(self._order1 + 1):
            if x1_ind > 0: # The first row should be 0-th order x1 partials.
                # Append one x1 derivative.
                next_deriv = _append_jvp(
                    self._eval_fun_derivs[x1_ind - 1][0],
                    num_base_args=2, argnum=0)
                self._eval_fun_derivs.append([ next_deriv ])
            for x2_ind in range(self._order2):
                # Append one x2 derivative.  Note that the array already contains
                # a 0-th order x2 partial, and we are appending order2 new
                # derivatives, for a max order of order2 in order2 + 1 columns.
                next_deriv = _append_jvp(
                    self._eval_fun_derivs[x1_ind][x2_ind],
                    num_base_args=2, argnum=1)
                self._eval_fun_derivs[x1_ind].append(next_deriv)

    def eval_directional_derivative(self, x1, x2, dx1s, dx2s, validate=True):
        """Evaluate the directional derivative in the directions
        dx1s[0] ..., dx1s[N_1], dx2s[0], ..., dx2s[N_2].
        """

        order1 = len(dx1s)
        order2 = len(dx2s)
        if validate:
            if order1 > self._order1:
                raise ValueError(
                    'The number of `dx1s` () must be <= order1 = {}'.format(
                        order1, self._order1))
            if order2 > self._order2:
                raise ValueError(
                    'The number of `dx2s` () must be <= order2 = {}'.format(
                        order1, self._order1))
            if (not isinstance(dx1s, list)) or (not isinstance(dx2s, list)):
                raise ValueError('`dx1s` and `dx2s` must be lists of vectors.')

        return self._eval_fun_derivs[order1][order2](x1, x2, *[*dx1s, *dx2s])


class ReverseModeDerivativeArray():
    def __init__(self, fun, order1, order2):
        self._order1 = order1
        self._order2 = order2

        # Build an array of higher order reverse mode partial derivatives.
        # Traverse the rows (x1 partials) first, then traverse the
        # columns (x2 partials).
        self._eval_deriv_arrays = [[ fun ]]
        for x1_ind in range(self._order1 + 1):
            if x1_ind > 0:
                # Append one x1 derivative.
                next_deriv = autograd.jacobian(
                    self._eval_deriv_arrays[x1_ind - 1][0], argnum=0)
                self._eval_deriv_arrays.append([ next_deriv ])
            for x2_ind in range(self._order2):
                # Append one x2 derivative.  Note that the array already has
                # a 0-th order x2 partial, and we are appending order2 new
                # derivatives, for a max order of order2 in order2 + 1 columns.
                next_deriv = autograd.jacobian(
                    self._eval_deriv_arrays[x1_ind][x2_ind], argnum=1)
                self._eval_deriv_arrays[x1_ind].append(next_deriv)

    def set_evaluation_location(self, x1, x2, force=False, verbose=False):
        """Set the location at which we evaluate the partial derivatives.
        """

        x1 = np.atleast_1d(x1)
        x2 = np.atleast_1d(x2)
        if x1.ndim != 1 or x2.ndim != 1:
            raise ValueError('x1 and x2 must be 1d arrays.')

        self._deriv_arrays = \
            [[ np.atleast_1d(self._eval_deriv_arrays[0][0](x1, x2)) ]]
        dim0 = len(self._deriv_arrays[0][0])
        dim1 = len(x1)
        dim2 = len(x2)
        total_size = 0
        for i1 in range(self._order1 + 1):
            for i2 in range(self._order2 + 1):
                total_size += dim0 * (dim1 ** i1) * (dim2 ** i2)

        max_allowed_size = 100000
        if total_size > max_allowed_size and not force:
            err_msg = ('With len(x1) = {}, len(x2) = {}, ' +
                       'order1 = {}, and order2 = {}, this will create a ' +
                       'partial derivative array of size {} > {}.  ' +
                       'To force the creation of these arrays, set' +
                       '`force=True`.').format(
                dim1, dim2, self._order1, self._order2,
                total_size, max_allowed_size)
            raise ValueError(err_msg)

        if self._order1 > 2 and self._order2 > 2 and not force:
            err_msg = ('With both orders greater than two, reverse ' +
                       'mode can be slow even in low-dimensional problems.  ' +
                       'To force the creation of the arrays, set force=True.')
            raise ValueError(err_msg)

        # Save the location at which the partial derivatives are evaluated.
        self._x1 = deepcopy(x1)
        self._x2 = deepcopy(x2)
        if self._deriv_arrays[0][0].ndim != 1:
            raise ValueError(
                'The base function is expected to evaluate to a 1d vector.')

        self._deriv_arrays = \
            [[ None for _ in range(self._order2 + 1)] \
                for _ in range(self._order1 + 1)]
        for x1_ind in range(self._order1 + 1):
            for x2_ind in range(self._order2 + 1):
                if verbose:
                    print('Evaluating the derivative {}, {}'.format(
                        x1_ind, x2_ind))
                self._deriv_arrays[x1_ind][x2_ind] = \
                    self._eval_deriv_arrays[x1_ind][x2_ind](x1, x2)

    def deriv_arrays(self, order1, order2):
        return self._deriv_arrays[order1][order2]

    def _check_location(self, x1, x2, tol=1e-8):
        if (np.max(np.abs(x1 - self._x1)) > tol) or \
            (np.max(np.abs(x2 - self._x2)) > tol):
            raise ValueError(
                'You must use the x1 and x2 set in `set_evaluation_location`.')

    def eval_directional_derivative(self, x1, x2, dx1s, dx2s, validate=True):
        """Evaluate the directional derivative in the directions
        dx1s[0] ..., dx1s[N_1], dx2s[0], ..., dx2s[N_2].
        """

        order1 = len(dx1s)
        order2 = len(dx2s)
        if validate:
            self._check_location(x1, x2)
            if order1 > self._order1:
                raise ValueError(
                    'The number of `dx1s` () must be <= order1 = {}'.format(
                        order1, self._order1))
            if order2 > self._order2:
                raise ValueError(
                    'The number of `dx2s` () must be <= order2 = {}'.format(
                        order1, self._order1))

        # This keeps the first dimension of the partial derivative array,
        # and sums over the x1 first, then the x2.  This needs to match the
        # order in which the partial derivatives are taken in `__init__`.
        return _contract_tensor(
            self._deriv_arrays[order1][order2], dx1s, dx2s)


class ReorderedReverseModeDerivativeArray():
    """
    Like a ReverseModeDerivativeArray, but takes derivatives with respect
    to the larger dimension last for efficiency.

    ReverseModeDerivativeArray takes derivatives with respect to the
    second variable last.  The second variable should have the largest
    dimension for efficient automatic differentiation.  This class wraps
    the methods of ReverseModeDerivativeArray, swapping the order of the
    arguments if necessary to evaluate derivatives in the more efficient
    direction.
    """
    def __init__(self, fun, order1, order2, swapped=None):
        self._swapped = swapped

        # The underlying ReverseModeDerivativeArray will be with respect to
        # z1, z2 for the purposes of naming variables.
        if self._swapped:
            self._orderz1 = order2
            self._orderz2 = order1
        else:
            self._orderz1 = order1
            self._orderz2 = order2

        def _fun(z1, z2):
            return fun(*(self._swapped_args(z1, z2)))
        self._fun = _fun

        self._rmda = ReverseModeDerivativeArray(
            fun=self._fun, order1=self._orderz1, order2=self._orderz2)

    def _swapped_args(self, x1, x2):
        if self._swapped:
            return x2, x1
        else:
            return x1, x2

    def set_evaluation_location(self, x1, x2, force=False, verbose=False):
        z1, z2 = self._swapped_args(x1, x2)
        return self._rmda.set_evaluation_location(
            x1=z1, x2=z2, force=force, verbose=verbose)

    def eval_directional_derivative(self, x1, x2, dx1s, dx2s, validate=True):
        z1, z2 = self._swapped_args(x1, x2)
        dz1s, dz2s = self._swapped_args(dx1s, dx2s)
        return self._rmda.eval_directional_derivative(
            z1, z2, dz1s, dz2s, validate=validate)

    def deriv_arrays(self, order1, order2):
        if self._swapped:
            orig_array = self._rmda.deriv_arrays(order2, order1)
            # Derivatives with respect to the second variable come first
            # (after the axis of the base function), and need to be at the end.
            axes2 = np.arange(order2) + 1
            return np.moveaxis(orig_array, axes2, -1 * axes2)
        else:
            return self._rmda.deriv_arrays(order1, order2)


def _consolidate_terms(dterms):
    """
    Combine like derivative terms.

    Arguments
    -----------
    dterms:
        A list of DerivativeTerms.

    Returns
    ------------
    A new list of derivative terms that evaluate equivalently where
    terms with the same derivative signature have been combined.
    """
    unmatched_indices = [ ind for ind in range(len(dterms)) ]
    consolidated_dterms = []
    while len(unmatched_indices) > 0:
        match_term = dterms[unmatched_indices.pop(0)]
        for ind in unmatched_indices:
            if (match_term.eta_orders == dterms[ind].eta_orders):
                match_term = match_term.combine_with(dterms[ind])
                unmatched_indices.remove(ind)
        consolidated_dterms.append(match_term)

    return consolidated_dterms


# Get the terms to start a Taylor expansion.
def _get_taylor_base_terms():
    dterms1 = [ \
        DerivativeTerm(
            eps_order=1,
            eta_orders=[0],
            prefactor=1.0),
        DerivativeTerm(
            eps_order=0,
            eta_orders=[1],
            prefactor=1.0) ]
    return dterms1


class ParametricSensitivityTaylorExpansion(object):
    """
    Evaluate the Taylor series of an optimum on a hyperparameter.

    This is a class for computing the Taylor series of
    eta(eps) = argmax_eta objective(eta, eps) using forward-mode automatic
    differentation.

    .. note:: This class is experimental and should be used with caution.
    """
    @classmethod
    def optimization_objective(
        cls,
        objective_function,
        input_val0,
        hyper_val0,
        order,
        hess0=None,
        forward_mode=True,
        max_input_order=None,
        max_hyper_order=None,
        force=False):
        """
        Parameters
        ------------------
        objective_function : `callable`
            The optimization objective as a function of two arguments
            (eta, eps), where eta is the parameter that is optimized and
            eps is a hyperparameter.
        hess0 : `numpy.ndarray` (N, N)
            Optional.  The Hessian of the objective at
            (``input_val0``, ``hyper_val0``).
            If not specified it is calculated at initialization.
        The remaining arguments are the same as for the `__init__` method.
        """
        # In order to calculate derivatives d^kinput_dhyper^k, we will be
        # Taylor expanding the gradient of the objective with respect to eta.
        estimating_equation = autograd.grad(objective_function, argnum=0)

        if hess0 is None:
            objective_function_hessian = \
                autograd.hessian(objective_function, argnum=0)
            hess0 = objective_function_hessian(input_val0, hyper_val0)

        # TODO: if the objective function returns a 1-d array and not a
        # float then the Cholesky decomposition will fail because
        # the Hessian will have an extra dimension.  This is a confusing
        # error that we could catch explicitly at the cost of an extra
        # function evaluation.  Is it worth it?
        hess_solver = solver_lib.get_cholesky_solver(hess0)

        return cls(
            estimating_equation=estimating_equation,
            input_val0=input_val0,
            hyper_val0=hyper_val0,
            order=order,
            hess_solver=hess_solver,
            forward_mode=forward_mode,
            max_input_order=max_input_order,
            max_hyper_order=max_hyper_order,
            force=force)

    def __init__(self,
                 estimating_equation,
                 input_val0, hyper_val0,
                 order,
                 hess_solver,
                 forward_mode=True,
                 max_input_order=None,
                 max_hyper_order=None,
                 force=False):
        """
        Parameters
        ------------------
        estimating_equation : `callable`
            A vector-valued function function of two arguments,
            (`input`, `output`), where the length of the vector is the same
            as the length of `input`, and which is (approximately) the zero
            vector when evaluated at (`input_val0`, `hyper_val0`).
        input_val0 : `numpy.ndarray` (N,)
            The value of ``input_par`` at the optimum.
        hyper_val0 : `numpy.ndarray` (M,)
            The value of ``hyper_par`` at which ``input_val0`` was found.
        order : `int`
            The maximum order of the Taylor series to be calculated.
        hess_solver : `function`
            A function that takes a single argument, `v`, and returns

            .. math::

                \\frac{\\partial G}{\\partial \\eta}^{-1} v,

            where :math:`G(\\eta, \\epsilon)` is the estimating equation,
            and the partial derivative is evaluated at
            :math:`(\\eta, \\epsilon) =`  (`input_val0`, `hyper_val0`).
        forward_mode : `bool`
            Optional.  If `True` (the default), use forward-mode automatic
            differentiation.  Otherwise, use reverse-mode.
        max_input_order : `int`
            Optional.  The maximum number of nonzero partial derivatives of
            the objective function gradient with respect to the input parameter.
            If `None`, calculate partial derivatives of all orders.
        max_hyper_order : `int`
            Optional.  The maximum number of nonzero partial derivatives of
            the objective function gradient with respect to the hyperparameter.
            If `None`, calculate partial derivatives of all orders.
        force: `bool`
            Optional.  If `True`, force the instantiation of potentially
            expensive reverse mode derivative arrays.  Default is `False`.
        """
        self._input_val0 = deepcopy(input_val0)
        self._hyper_val0 = deepcopy(hyper_val0)
        self._objective_function_eta_grad = estimating_equation
        self._set_order(order, max_input_order, max_hyper_order, forward_mode)
        self.hess_solver = hess_solver

        if not self._forward_mode:
            # Set the derivative arrays.
            # TODO: don't duplicate the Hessian calculation?
            self._deriv_array.set_evaluation_location(
                self._input_val0, self._hyper_val0, force=force)

    def _set_order(self, order, max_input_order, max_hyper_order, forward_mode):
        """Generate the matrix of g partial derivatives and differentiate
        the Taylor series up to the required order.
        """
        self._max_input_order = max_input_order
        self._max_hyper_order = max_hyper_order
        self._forward_mode = forward_mode
        if not self._forward_mode:
            warnings.warn(
                'Reverse mode Taylor expansions are experimental.')

        if self._max_input_order is not None or \
           self._max_hyper_order is not None:
            warnings.warn(
                'Setting _max_hyper_order or _max_input_order is experimental.')

        if self._max_input_order is not None and self._max_input_order < 1:
            raise ValueError('max_input_order must be >= 1.')
        if self._max_hyper_order is not None and self._max_hyper_order < 1:
            raise ValueError('max_hyper_order must be >= 1.')

        self._order = order

        # In fact can't this be self._order - 1?
        if self._max_input_order is None:
            order1 = self._order
        else:
            order1 = min(self._order, self._max_input_order)

        if self._max_hyper_order is None:
            order2 = self._order
        else:
            order2 = min(self._order, self._max_hyper_order)

        if self._forward_mode:
            self._deriv_array = ForwardModeDerivativeArray(
                self._objective_function_eta_grad,
                order1=order1,
                order2=order2)
        else:
            # If the input has higher dimension than the hyperparameter,
            # we want to take reverse mode derivatives with respect to
            # the hyperparameter first, so swap the order of the arguments.
            swapped = len(self._input_val0) > len(self._hyper_val0)
            self._deriv_array = ReorderedReverseModeDerivativeArray(
                self._objective_function_eta_grad,
                order1=order1,
                order2=order2,
                swapped=swapped)

        def differentiate_terms(dterms):
            dterms_derivs = []
            for term in dterms:
                dterms_derivs += term.differentiate()
            return _consolidate_terms(dterms_derivs)

        self._taylor_terms_list = [ _get_taylor_base_terms() ]
        for k in range(1, self._order):
            next_taylor_terms = \
                differentiate_terms(self._taylor_terms_list[k - 1])
            self._taylor_terms_list.append(next_taylor_terms)

    def get_max_order(self):
        return self._order

    def _evaluate_dkinput_dhyperk(self, dhyper, input_derivs, k):
        """
        Evaluate the derivative d^k input / d hyper^k in the direction dhyper.

        Parameters
        --------------
        dhyper : `numpy.ndarray` (N, )
            The direction ``new_hyper_val - hyper_val0`` in which to evaluate
            the directional derivative.
        input_derivs : `list` of `numpy.ndarray`
            A list of previous dkinput_dhyperk up to order k - 1.
        k : `int`
            The order of the derivative.

        Returns
        ------------
            The value of the k^th derivative in the direction dhyper.
        """
        if k <= 0:
            raise ValueError('k must be at least one.')
        if k > self._order:
            raise ValueError(
                'k must be no greater than the declared order={}'.format(
                    self._order))
        if len(input_derivs) < k - 1:
            raise ValueError('Not enough eta_derivs provided.')
        vec = np.zeros_like(self._input_val0)
        for term in self._taylor_terms_list[k - 1]:
            # Exclude the highest order eta derivative -- this what
            # we are trying to calculate.
            if (term.eta_orders[-1] > 0):
                continue

            # Assume that epsilon partials higher than this are zero.
            if self._max_hyper_order is not None and \
               term.eps_order > self._max_hyper_order:
                continue

            # Assume that eta partials higher than this are zero.
            if self._max_input_order is not None and \
               term.total_eta_order > self._max_input_order:
                continue

            vec += \
                _evaluate_term_fwd(
                    term=term,
                    eta0=self._input_val0,
                    eps0=self._hyper_val0,
                    deps=dhyper,
                    eta_derivs=input_derivs,
                    eval_directional_derivative= \
                        self._deriv_array.eval_directional_derivative)
        return -1 * self.hess_solver(vec)


    def _get_default_max_order(self, max_order):
        if max_order is None:
            return self._order
        if max_order <= 0:
            raise ValueError('max_order must be greater than zero.')
        if max_order > self._order:
            raise ValueError(
                'max_order must be no greater than the order={}'.format(
                    self._order))
        return max_order

    def evaluate_input_derivs(self, dhyper, max_order=None):
        """Return a list of the derivatives dkinput / dhyperk dhyper^k
        """
        max_order = self._get_default_max_order(max_order)
        input_derivs = []
        for k in range(1, max_order + 1):
            dinputk_dhyperk = \
                self._evaluate_dkinput_dhyperk(
                    dhyper=dhyper,
                    input_derivs=input_derivs,
                    k=k)
            input_derivs.append(dinputk_dhyperk)
        return input_derivs


    def evaluate_taylor_series_terms(self, new_hyper_val, add_offset=True,
                                     max_order=None):
        """Return the terms in a Taylor series approximation.
        """
        max_order = self._get_default_max_order(max_order)
        if add_offset:
            dinput_terms = [self._input_val0]
        else:
            dinput_terms = [np.zeros_like(self._input_val0)]
        dhyper = new_hyper_val - self._hyper_val0
        input_derivs = self.evaluate_input_derivs(dhyper, max_order=max_order)

        for k in range(1, max_order + 1):
            dinput_terms.append(input_derivs[k - 1] / float(factorial(k)))

        return dinput_terms


    def evaluate_taylor_series(self, new_hyper_val,
                               add_offset=True, max_order=None,
                               sum_terms=True):
        """
        Evaluate the derivative ``d^k input / d hyper^k`` in the direction dhyper.

        Parameters
        --------------
        new_hyper_val: `numpy.ndarray` (N, )
            The new hyperparameter value at which to evaluate the
            Taylor series.
        add_offset: `bool`
            Optional.  Whether to add the initial constant input_val0 to the
            Taylor series.
        max_order: `int`
            Optional.  The order of the Taylor series.  Defaults to the
            ``order`` argument to ``__init__``.
        sum_terms: `bool`
            If ``True``, add the terms in the Taylor series.  If ``False``,
            return the terms as a list.

        Returns
        ------------
            The Taylor series approximation to ``input_val(new_hyper_val)`` if
            ``add_offset`` is ``True``, or to
            ``input_val(new_hyper_val) - input_val0`` if ``False``.  If
            ``sum_terms`` is ``True``, then a vector of the same length as
            ``input_val`` is returned.  Otherwise, an array of
            shape ``max_order + 1, len(input_val)`` is returned containing
            the terms of the Taylor series approximation.
        """

        dinput_terms = self.evaluate_taylor_series_terms(
            new_hyper_val=new_hyper_val,
            add_offset=add_offset,
            max_order=max_order)
        return np.sum(dinput_terms, axis=0)


    def print_terms(self, k=None):
        """
        Print the derivative terms in the Taylor series.

        Parameters
        ---------------
        k: integer
            Optional.  Which term to print.  If unspecified, all terms are
            printed.
        """
        if k is not None and k > self._order:
            raise ValueError(
                'k must be no greater than order={}'.format(self._order))
        for order in range(self._order):
            if k is None or order == (k - 1):
                print('\nTerms for order {}:'.format(order + 1))
                for term in self._taylor_terms_list[order]:
                    print(term)



import autograd
import numpy as np
from .sensitivity_lib import _append_jvp

from copy import deepcopy
import scipy as sp
import scipy.sparse
from scipy.sparse import coo_matrix


class SparseBlockHessian():
    """Efficiently calculate block-sparse Hessians.

        The objective function is expected to be of the form

        .. math ::
            x = (x_1 , ... , x_G) \\textrm{ (or some permutation thereof)}

            f(x) = \sum_{g=1}^{G} f_g(x_g)

        Each :math:`x_g` is
        expected to have the same dimension.  Consequently, the Hessian
        matrix of ``f`` with respect to ``x``, is block diagonal with
        ``G`` blocks, up to a permutation of the order of ``x``.
        The purpose of this class is to efficiently calculate
        this Hessian when the block structure (i.e., the partition of ``x``)
        is known.

    """
    def __init__(self, objective_function, sparsity_array):
        """In terms of the class description, ``objective_function = f``,
        ``opt_par = x``, and ``sparsity_array`` describes
        the partition of ``x`` into :math:`(x_1, ..., x_G)`.

        Parameters
        ------------
        objective_function : `callable`
            An objective function of which to calculate a Hessian.   The
            argument should be

            - ``opt_par``: `numpy.ndarray` (N,) The optimization parameter.

        sparsity_array : `numpy.ndarray` (G, M)
            An array containing the indices of rows and columns of each block.
            The Hessian should contain ``G`` dense blocks, each of which
            is ``M`` by ``M``.  Each row of ``sparsity_array`` should contain
            the indices of the corresponding block.  There must be no repeated
            indices, and each block must be the same size.
        """
        self._fun = objective_function
        self._sparsity_array = sparsity_array
        self._num_blocks = self._sparsity_array.shape[0]
        self._block_size = self._sparsity_array.shape[1]

        if len(np.unique(sparsity_array)) != len(sparsity_array.flatten()):
            raise ValueError(
                'The indices in ``sparsity array`` must be unique.')

        self._f_grad = autograd.grad(self._fun, argnum=0)
        self._f_fwd_hess = _append_jvp(self._f_grad, num_base_args=1)

    def _hess_summed_term(self, opt_par, ib):
        """``ib`` is the index within the block.
        """
        v = np.zeros_like(opt_par)
        v[self._sparsity_array[:, ib]] = 1
        return self._f_fwd_hess(opt_par, v)

    def get_block_hessian(self, opt_par, print_every=0):
        """Get the block Hessian at ``opt_par`` and ``weights``.

        Parmeters
        ----------
        opt_par : `numpy.ndarray`
            The argument to ``objective_function`` at which to evaluate
            the Hessian matrix.
        print_every : `int`, optional.
            How often to display progress.  If ``0``, nothing is printed.

        Returns
        --------
        hessian : `scipy.sparse.coo_matrix` (N, N)
            The block-sparse Hessian given by and ``sparsity_array``.
        """
        opt_par = np.atleast_1d(opt_par)
        if opt_par.ndim != 1:
            raise ValueError('``opt_par`` must be a vector.')

        mat_vals = [] # These will be the entries of the Hessian
        mat_rows = [] # These will be the row indices
        mat_cols = [] # These will be the column indices

        for ib in range(self._block_size):
            if print_every > 0:
                if ib % print_every == 0:
                    print('Block index {} of {}.'.format(ib, self._block_size))
            hess_prod = self._hess_summed_term(opt_par, ib)
            for b in range(self._num_blocks):
                hess_inds = self._sparsity_array[b, :]
                mat_vals.extend(hess_prod[hess_inds])
                mat_rows.extend(hess_inds)
                mat_cols.extend(np.full(self._block_size, hess_inds[ib]))
        if print_every > 0:
            print('Done differentiating.')

        d = len(opt_par)
        h_sparse = coo_matrix((mat_vals, (mat_rows, mat_cols)), (d, d))
        return h_sparse

    def get_global_hessian(self, opt_par, global_inds=None, print_every=0):
        """Get the dense Hessian terms for the global parameters, which
        are, by default, indexed by any indices not in ``_sparsity_array``.
        """
        local_inds = np.hstack(self._sparsity_array)
        if global_inds is None:
            global_inds = np.setdiff1d(np.arange(len(opt_par)), local_inds)

        global_local_intersection = np.intersect1d(global_inds, local_inds)
        if len(global_local_intersection) > 0:
            raise ValueError(
                'The global and local indices must be disjoint.  {}'.format(
                    global_local_intersection))

        mat_vals = [] # These will be the entries of the Hessian
        mat_rows = [] # These will be the row indices
        mat_cols = [] # These will be the column indices

        v = np.zeros_like(opt_par)
        count = 0
        for ig in global_inds:
            if print_every > 0:
                if count % print_every == 0:
                    print('Global index {} of {}.'.format(
                        count, len(global_inds)))
            v[ig] = 1
            hess_row = self._f_fwd_hess(opt_par, v)
            for il in local_inds:
                mat_vals.append(hess_row[il])
                mat_cols.append(ig)
                mat_rows.append(il)

                mat_vals.append(hess_row[il])
                mat_cols.append(il)
                mat_rows.append(ig)

            for ig2 in global_inds:
                mat_vals.append(0.5 * hess_row[ig2])
                mat_cols.append(ig)
                mat_rows.append(ig2)

                mat_vals.append(0.5 * hess_row[ig2])
                mat_cols.append(ig2)
                mat_rows.append(ig)

            v[ig] = 0
            count += 1

        if print_every > 0:
            print('Done differentiating.')

        d = len(opt_par)
        h_sparse = coo_matrix((mat_vals, (mat_rows, mat_cols)), (d, d))
        return h_sparse

    def get_hessian(self, opt_par, print_every=0):
        local_hessian = self.get_block_hessian(opt_par, print_every=print_every)
        global_hessian = self.get_global_hessian(opt_par, print_every=print_every)
        return local_hessian + global_hessian
